#include "Encoding.hpp"

#include "Error.hpp"

#include <stdio.h>
#include <string.h>

namespace SoftWire
{
	Encoding::Encoding()
	{
		reset();

		O1 = 0xCC;   // Breakpoint
		format.O1 = true;
	}

	Encoding::~Encoding()
	{
		delete[] label;
		delete[] reference;
	}

	void Encoding::reset()
	{
		label = 0;
		reference = 0;

		format.P1 = false;
		format.P2 = false;
		format.P3 = false;
		format.P4 = false;
		format.O2 = false;
		format.O1 = false;
		format.modRM = false;
		format.SIB = false;
		format.D1 = false;
		format.D2 = false;
		format.D3 = false;
		format.D4 = false;
		format.I1 = false;
		format.I2 = false;
		format.I3 = false;
		format.I4 = false;

		P1 = 0;
		P2 = 0;
		P3 = 0;
		P4 = 0;
		O2 = 0;
		O1 = 0;
		modRM.b = 0;
		SIB.b = 0;
		D1 = 0;
		D2 = 0;
		D3 = 0;
		D4 = 0;
		I1 = 0;
		I2 = 0;
		I3 = 0;
		I4 = 0;

		immediate = 0;
		displacement = 0;
	}

	const char *Encoding::getLabel() const
	{
		return label;
	}

	const char *Encoding::getReference() const
	{
		return reference;
	}

	int Encoding::length() const
	{
		int length = 0;

		if(format.P1)		length++;
		if(format.P2)		length++;
		if(format.P3)		length++;
		if(format.P4)		length++;
		if(format.O2)		length++;
		if(format.O1)		length++;
		if(format.modRM)	length++;
		if(format.SIB)		length++;
		if(format.D1)		length++;
		if(format.D2)		length++;
		if(format.D3)		length++;
		if(format.D4)		length++;
		if(format.I1)		length++;
		if(format.I2)		length++;
		if(format.I3)		length++;
		if(format.I4)		length++;

		return length;
	}

	void Encoding::addPrefix(byte p)
	{
		if(!format.P1)
		{
			P1 = p;
			format.P1 = true;;
		}
		else if(!format.P2)
		{
			P2 = p;
			format.P2 = true;
		}
		else if(!format.P3)
		{
			P3 = p;
			format.P3 = true;
		}
		else if(!format.P4)
		{
			P4 = p;
			format.P4 = true;
		}
		else
		{
			throw Error("Too many prefixes in opcode");
		}
	}

	void Encoding::setImmediate(int immediate)
	{
		this->immediate = immediate;
	}

	void Encoding::setJumpOffset(int offset)
	{
		if((char)offset != offset && format.I2)
		{
			throw Error("Jump offset range too big");
		}

		this->immediate = offset;
	}

	void Encoding::setLabel(const char *label)
	{
		this->label = strdup(label);
	}

	void Encoding::setReference(const char *label)
	{
		reference = strdup(label);
	}

	void Encoding::checkFormat() const
	{
		// Bytes cannot be changed without updating format, except immediate and displacement
		if((P1 && !format.P1) ||
		   (P2 && !format.P2) ||
		   (P3 && !format.P3) ||
		   (P4 && !format.P4) ||
		   (O2 && !format.O2) ||
		   (O1 && !format.O1) ||
		   (modRM.b && !format.modRM) ||
		   (SIB.b && !format.SIB))
		{
			throw Error::INTERNAL;   
		}

		if((format.P4 && !format.P3) ||
		   (format.P3 && !format.P2) ||
		   (format.P2 && !format.P1))
		{
			throw Error::INTERNAL;
		}

		if(format.O2 &&
		   (O2 != 0x0F &&
		    O2 != 0xD8 &&
			O2 != 0xD9 &&
			O2 != 0xDA &&
			O2 != 0xDB &&
			O2 != 0xDC &&
			O2 != 0xDD &&
			O2 != 0xDE &&
			O2 != 0xDF))
		{
			throw Error::INTERNAL;
		}

		if(format.SIB)
		{
			if(!format.modRM)
			{
				throw Error::INTERNAL;
			}

			if(modRM.r_m != ESP)
			{
				throw Error::INTERNAL;
			}
		}

		// Byte, word or doubleword
		if((format.D4 && !format.D3) ||
		   (format.D3 && !format.D4) ||
		   (format.D3 && !format.D2) ||
		   (format.D2 && !format.D1))
		{
			throw Error::INTERNAL;
		}

		// Byte, word or doubleword
		if((format.I4 && !format.I3) ||
		   (format.I3 && !format.I4) ||
		   (format.I3 && !format.I2) ||
		   (format.I2 && !format.I1))
		{
			throw Error::INTERNAL;
		}
	}

	int Encoding::writeCode(byte *output) const
	{
		#ifndef NDEBUG
			checkFormat();

			if(format.P1)		printf("%.2X ", P1);
			if(format.P2)		printf("%.2X ", P2);
			if(format.P3)		printf("%.2X ", P3);
			if(format.P4)		printf("%.2X ", P4);
			if(format.O2)		printf("%.2X ", O2);
			if(format.O1)		printf("%.2X ", O1);
			if(format.modRM)	printf("%.2X ", modRM.b);
			if(format.SIB)		printf("%.2X ", SIB.b);
			if(format.D1)		printf("%.2X ", D1);
			if(format.D2)		printf("%.2X ", D2);
			if(format.D3)		printf("%.2X ", D3);
			if(format.D4)		printf("%.2X ", D4);
			if(format.I1)		printf("%.2X ", I1);
			if(format.I2)		printf("%.2X ", I2);
			if(format.I3)		printf("%.2X ", I3);
			if(format.I4)		printf("%.2X ", I4);

			printf("\n");
		#endif

		byte *start = output;

		if(format.P1)		*output++ = P1;
		if(format.P2)		*output++ = P2;
		if(format.P3)		*output++ = P3;
		if(format.P4)		*output++ = P4;
		if(format.O2)		*output++ = O2;
		if(format.O1)		*output++ = O1;
		if(format.modRM)	*output++ = modRM.b;
		if(format.SIB)		*output++ = SIB.b;
		if(format.D1)		*output++ = D1;
		if(format.D2)		*output++ = D2;
		if(format.D3)		*output++ = D3;
		if(format.D4)		*output++ = D4;
		if(format.I1)		*output++ = I1;
		if(format.I2)		*output++ = I2;
		if(format.I3)		*output++ = I3;
		if(format.I4)		*output++ = I4;

		return output - start;
	}
}