#include "Scanner.hpp"

#include "Token.hpp"
#include "Error.hpp"

#include <ctype.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

namespace SoftWire
{
	Scanner::Link::Link()
	{
		token = 0;
		next = 0;
	}

	Scanner::Link::~Link()
	{
		delete token;
		delete next;
	}

	Scanner::Scanner()
	{
		root = 0;
		pointer = 0;
	}

	Scanner::~Scanner()
	{
		delete root;
	}

	void Scanner::setSource(const char *source)
	{
		if(!source)
		{
			throw Error::INTERNAL;
		}

		delete root;
		root = new Link();

		Link *link = root;

		int i = 0;

		while(link)
		{
			char buffer[tokenMax];
			int c = 0;

			switch(source[i])
			{
			case '\0':
			case '\n':
			case '\r':
			case ';':
				link->token = new Token();
				pointer = root;
				return;
			case ' ':
			case '	':
				i++;
				continue;
			case '\'':
				throw Error("Character constants not supported yet");
			case '\"':
				throw Error("String literals not supported yet");
			case '*':
			case '+':
			case '-':
			case '~':
			case ',':
			case '[':
			case ']':
			case ':':
				link->token = new Punctuator(source[i++]);
				break;
			default:
				if(__iscsymf(source[i]))
				{
					do
					{
						if(c >= tokenMax)
						{
							throw Error("Token too long");
						}

						buffer[c++] = source[i++];
					}
					while(__iscsym(source[i]) && c < tokenMax);

					buffer[c] = 0;

					link->token = new Identifier(buffer);
				}
				else if(isdigit(source[i]))
				{
					do
					{
						buffer[c++] = source[i++];
					}
					while((isxdigit(source[i]) || source[i] == 'x') && c < tokenMax - 1);

					buffer[c] = 0;
					char *endptr;

					if(source[i] == 'h')   // Hexadecimal
					{
						link->token = new Constant(strtoul(buffer, &endptr, 16));

						buffer[c++] = source[i++];
					}
					else if(source[i] == 'q')   // Octal
					{
						link->token = new Constant(strtoul(buffer, &endptr, 8));

						buffer[c++] = source[i++];
					}
					else if(source[i] == 'b')   // Binary
					{
						link->token = new Constant(strtoul(buffer, &endptr, 2));

						buffer[c++] = source[i++];
					}
					else   // Decimal
					{
						link->token = new Constant(strtol(buffer, &endptr, 10));
					}
				}
				else
				{
					throw Error("Invalid token '%c'", source[i]);
				}
			}

			link->next = new Link();
			link = link->next;
		}

		throw Error::INTERNAL;

		pointer = root;

		return;
	}

	bool Scanner::isEndOfLine() const
	{
		assertPointer();

		return pointer->token->isEndOfLine();
	}

	bool Scanner::isIdentifier() const
	{
		assertPointer();

		return pointer->token->isIdentifier();
	}

	bool Scanner::isConstant() const
	{
		assertPointer();

		return pointer->token->isConstant();
	}

	bool Scanner::isPunctuator() const
	{
		assertPointer();

		return pointer->token->isPunctuator();
	}

	const Token &Scanner::advance(int n)
	{
		while(n--)
		{
			assertPointer();

			pointer = pointer->next;
		}

		return *pointer->token;
	}

	const char *Scanner::getString() const
	{
		assertPointer();

		return pointer->token->getString();
	}

	const Token &Scanner::lookAhead() const
	{
		assertPointer();

		return *pointer->next->token;
	}

	char Scanner::getChar() const
	{
		assertPointer();

		return pointer->token->getChar();
	}

	int Scanner::getValue() const
	{
		assertPointer();

		return pointer->token->getValue();
	}

	const Token &Scanner::currentToken() const
	{
		assertPointer();

		return *pointer->token;
	}

	void Scanner::assertPointer() const
	{
		if(!pointer || !pointer->token)
		{
			throw Error::INTERNAL;
		}
	}
}